/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * sandnet.c
 * Part of libsandnet
 * Copyright (C) 2005 Simon Newton
 */

// required for vasprintf
// patches welcomed to remove it :)
#define _GNU_SOURCE

#include "private.h"
#include <stdio.h>
#include <stdarg.h>

char *sandnet_errstr = NULL;

/*
 * libsandnet error function
 *
 * This writes the error string to sandnet_strerror, which can be accessed
 * using sandnet_strerror();
 *
 * Is vasprintf available on OSX ?
 */
void sandnet_error(const char *fmt, ...) {

  va_list ap;
  va_start(ap, fmt);

  free(sandnet_errstr);
  vasprintf(&sandnet_errstr, fmt , ap) ,
  va_end(ap);

}


/*
 * Fetches the sandnet error string
 *
 * @return the error string
 */
char *sandnet_strerror() {
  return sandnet_errstr;
}

/**
 * Creates a new SandNet node.
 * Takes a string containing the ip address of the interface to listen on
 * If the string is NULL it uses the first non loopback address
 *
 * @param ip the IP address to listen on
 * @param debug level of logging provided 0: none
 * @return an sandnet_node
 */
sandnet_node sandnet_new(const char *ip, int debug) {
  node n;
  int ret,i;

  n = malloc(sizeof(sandnet_node_t ) );

  if(n == NULL) {
    sandnet_error("malloc failure");
    return NULL;
  }

  memset(n, 0x0, sizeof(sandnet_node_t) );

  // default parameters
  n->state.mode = SANDNET_OFF;
  n->state.verbose = debug;
  n->state.packet_count = 0;

  strncpy(n->state.name,"libsandnet", SANDNET_NAME_LENGTH);

  n->state.led = 0x00;
  // setup our ports
  for(i=0; i < SANDNET_MAX_PORTS; i++) {
    n->ports[i].proto = 0x02;      // sandnet dmx
    n->ports[i].mode = SANDNET_PORT_MODE_DISABLED;
    n->ports[i].term = 0x00;
    n->ports[i].b = 0x00;
    n->ports[i].param1 = 0x00;
    n->ports[i].param2 = i;
  }

  ret = sandnet_net_init(n, ip);

  if (ret) {
    free(n);
    return NULL;
  }

  return (sandnet_node) n;
}

/**
 * Starts the SandNet node.
 * Binds to the network sockets
 *
 * @param vn the sandnet_node
 * @return 0 on success, non 0 on failure
 *
 */
int sandnet_start(sandnet_node vn) {
  node n = (node) vn;
  int ret;

  return_if_null(vn);
  return_if_on(n);

  // build ad
  sandnet_tx_build_ad(n);

  if ((ret = sandnet_net_start(n)) ) {
    return ret;
  }
  n->state.mode = SANDNET_ON;

  return sandnet_send_ad(n);
}


/**
 * Stops the SandNet node
 * Closes the network sockets held by the node
 *
 * @param vn the sandnet_node
 * @return 0 on success, non 0 on failure
 */
int sandnet_stop(sandnet_node vn) {
  node n = (node) vn;
  int ret;

  return_if_null(vn);
  return_if_off(n);

  if ((ret = sandnet_net_close(n) ))
    return ret;

  n->state.mode = SANDNET_OFF;

  return SANDNET_EOK;
}


/**
 * Free the memory associated with this node
 *
 * @param vn the sandnet_node
 */
int sandnet_destroy(sandnet_node vn) {

  return_if_null(vn);

  free(vn);
  return SANDNET_EOK;
}


/**
 * Handle any received packets.
 * This function is the workhorse of libsandnet. You have a couple of options:
 *   - use sandnet_get_sd() to retrieve the socket descriptors and select to
 *     detect network activity. Then call sandnet_read(node,0)
 *     when activity is detected.
 *   - call sandnet_read repeatedly from within a loop with an appropriate
 *     timeout
 *
 * @param vn the sandnet_node
 * @param timeout the number of seconds to block for if nothing is pending
 * @return 0 on success, -1 on failure
 */
int sandnet_read(sandnet_node vn, int timeout) {
  node n = (node) vn;
  sandnet_packet_t p;
  int ret;

  return_if_null(vn);
  return_if_off(n);

  while(1) {
    memset( &p.data, 0x0, sizeof(p.data) );

    if( (ret = sandnet_net_recv(n, &p, timeout) ) < 0)
      return ret;

    // nothing to read
    if (ret == RECV_NO_DATA)
      break;

    // skip this packet (filtered)
    if (p.length == 0)
      continue;

    // should we be returning errors here?
    if( p.to.s_addr == n->state.ctl_addr.s_addr)
      sandnet_rx_handle_ctl(n, &p);
    else if (p.to.s_addr == n->state.data_addr.s_addr)
      sandnet_rx_handle_data(n,&p);

  }
  return SANDNET_EOK;
}




/**
 * This is used to set handlers for sent/received sandnet packets.
 * If you're using a stock standard node you more than likely don't want
 * to set these. See the sandnet_set_dmx_callback and sandnet_set_firmware_callback .
 * If you want to get down and dirty with sandnet packets, you can set this
 * read / manipulate packets as they arrive (or get sent)
 *
 * @param vn     The sandnet_node
 * @param handler  The handler to set
 * @param fh    A pointer to a function, set to NULL to turn off
 *             The function should return 0,
 * @param data    Data to be passed to the handler when its called
 *
 * @return 0 on sucess, non 0 on failure
 *
 */
int sandnet_set_handler(sandnet_node vn, sandnet_handler_name_t handler, int (*fh)(sandnet_node n, void *pp, void * d), void *data) {
  node n = (node) vn;

  return_if_null(vn);

  if(handler < 0 || handler >= SANDNET_LAST_HANDLER ) {
    sandnet_error("%s : Invalid handler defined", __FUNCTION__ );
    return SANDNET_EARG;
  }

  n->callbacks[handler].fh = fh;
  n->callbacks[handler].data = data;

  return SANDNET_EOK;
}


/*
 * This is a special callback which is invoked when dmx data is received.
 * Unlike libsandnet we pass the data back here, an esp node can recv on any
 * of 256 channels, we don't want to maintain buffer for them.
 *
 * @param vn     The sandnet_node
 * @param fh    The callback to invoke (parameters passwd are the sandnet_node, the port_id
 *           that received the dmx, and some user data
 * @param data    Data to be passed to the handler when its called
 *
 */
int sandnet_set_dmx_handler(sandnet_node vn, int (*fh)(sandnet_node n, uint8_t group, uint8_t universe, int length, uint8_t *data,  void *d), void *data ) {
  node n = (node) vn;

  return_if_null(vn);

  n->dmx_callback.fh = fh;
  n->dmx_callback.data = data;

  return SANDNET_EOK;
}



/*
 * Send a sandnet advertisment
 * This should be called every 2 seconds
 *
 */
int sandnet_send_ad(sandnet_node vn) {
  node n = (node) vn;

  return_if_null(vn);
  return_if_off(n);
  return sandnet_tx_send_ad(n);
}



/*
 * Sends some dmx data
 *
 * @param vn the sandnet_node
 * @param universe universe to address it to
 * @param length length of data
 * @param data pointer to dmx data
 *
 */
int sandnet_send_dmx(sandnet_node vn, int port, int16_t length, const uint8_t *data) {
  node n = (node) vn;

  return_if_null(vn);
  return_if_off(n);

  if( length < 1 || length > SANDNET_DMX_LENGTH) {
    sandnet_error("%s : Length of dmx data out of bounds (should be between 1 and 512, was %i)", __FUNCTION__ , length);
    return SANDNET_EARG;
  }

  if( port < 0 || port >= SANDNET_MAX_PORTS) {
    sandnet_error("%s : port index out of bounds (should be between 0 and 1, was %i)", __FUNCTION__ , port);
    return SANDNET_EARG;
  }

  if(n->ports[port].mode == SANDNET_PORT_MODE_IN) {
    return sandnet_tx_send_dmx(n, port, n->ports[port].param1, n->ports[port].param2, length, data);
  } else {
    sandnet_error("%s : attempt to send from non-input port %i", __FUNCTION__ , port);
    return SANDNET_ESTATE;

  }
}



/*
 * Sends some dmx data
 *
 * @param vn the sandnet_node
 * @param universe universe to address it to
 * @param length length of data
 * @param data pointer to dmx data
 *
 */
int sandnet_raw_send_dmx(sandnet_node vn, uint8_t port, uint8_t grp, uint8_t uni, int16_t length, const uint8_t *data) {
  node n = (node) vn;

  return_if_null(vn);
  return_if_off(n);

  if( length < 1 || length > SANDNET_DMX_LENGTH) {
    sandnet_error("%s : Length of dmx data out of bounds (should be between 1 and 512, was %i)", __FUNCTION__ , length);
    return SANDNET_EARG;
  }

  return sandnet_tx_send_dmx(n, port, grp, uni, length, data);
}


/**
 * Sets the name of the node.
 * The string should be null terminated and a maxmium of 18 Characters will be used
 *
 * @param vn the sandnet_node
 * @param name the short name of the node.
 */
int sandnet_set_name(sandnet_node vn, const char *name) {
  node n = (node) vn;

  return_if_null(vn);

  strncpy((char *) &n->state.name, name, SANDNET_NAME_LENGTH);
  return SANDNET_EOK;

}

/*
 * configure a port
 *
 * @param vn   the sandnet_node
 * @param port  the port index (0 or 1)
 * @param mode  the port mode sandnet_port_mode_t
 * @param param1  the first parameter
 * @param param2  the second parameter
 */
int sandnet_set_port(sandnet_node vn, int port, sandnet_port_mode_t mode, uint8_t param1, uint8_t param2) {
  node n = (node) vn;
  return_if_null(vn);

  if( port < 0 || port >= SANDNET_MAX_PORTS) {
    sandnet_error("%s : port index out of bounds (should be between 0 and 1, was %i)", __FUNCTION__ , port);
    return SANDNET_EARG;
  }

  n->ports[port].mode = mode;
  n->ports[port].param1 = param1;
  n->ports[port].param2 = param2;

  sandnet_tx_build_ad(n);
  sandnet_tx_send_ad(n);
  return SANDNET_EOK;
}



/**
 * Returns the socket descriptor associated with this sandnet_node.
 * libsandnet now uses a single descriptor per node  bound to the wildcard
 * address
 *
 * @param vn the sandnet_node
 * @param socket the index of the socket descriptor to fetch (0 or 1)
 * @return the socket descriptor, or < 0 on error
 */
int sandnet_get_sd(sandnet_node vn, int i) {
  node n = (node) vn;

  return_if_null(vn);
  return_if_off(n);

  if( i )
    return n->sd[1];
  else
    return n->sd[0];
}
